"""
All data-processing code methods for images, including profile extraction and masking procedures.
"""

from tqdm import tqdm
from astropy.table import Table
from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma, sigma_clipped_stats
from numpy import copy, ndarray, floor, nan, sum, ndarray, nanmean, nanmedian, nanstd
from photutils import detect_threshold, detect_sources, deblend_sources, make_source_mask
from photutils.aperture import CircularAnnulus, EllipticalAnnulus


def mask_cutout(cutout, config=None, nsigma=1., gauss_width=2.0, npixels=11, omit_centre=True,
                clip_negatives=True):
    """ Masks a cutout. Users can specify parameters to adjust the severity of the mask. Default
        parameters strikes a decent balance.

    Args:
        cutout: Input cutout to mask.
        config Input config file
        nsigma: The brightness requirement for objects.
        gauss_width: The width of the gaussian kernel.
        npixels: The minimum number of pixels that an object must be comprised of to be considered a source.
        omit_centre: Set as true to leave the central object unmasked.
        clip_negatives: Remove negative pixels that are 3 sigma below the median BG level
                           (This is to get rid of artifacts).
    Returns:
        The masked cutout and the associated mask data in tuple format.
    """

    # If provided with a config file, set the parameters to what is given by the user.
    # Note that the user can just use the regular specified parameters by keeping config as None
    if config is not None:
        params = config["MASK_PARAMS"]
        nsigma = params[0]
        gauss_width = params[1]
        npixels = params[2]

    mask_data = {}
    c_x, c_y = int(floor(cutout.shape[0] / 2)), int(floor(cutout.shape[1] / 2))

    # Generate background mask and statistics

    bg_mask = make_source_mask(cutout, nsigma=2, npixels=3, dilate_size=7)


    bg_mean, bg_median, bg_std = sigma_clipped_stats(cutout, sigma=3.0, mask=bg_mask)

    # Generate source mask
    source_mask = generate_mask(cutout, nsigma=nsigma, gauss_width=gauss_width, npixels=npixels)
    source_mask = boolean_mask(source_mask, omit=[source_mask[c_x][c_y]] if omit_centre else None)
    n_masked = sum(source_mask)

    masked_cutout = copy(cutout)
    masked_cutout[source_mask] = nan

    if clip_negatives:
        limit = bg_median - (3 * bg_std)
        masked_cutout[masked_cutout <= limit] = nan

    mask_data["BG_MEAN"] = bg_mean
    mask_data["BG_MEDIAN"] = bg_median
    mask_data["BG_STD"] = bg_std
    mask_data["N_MASKED"] = n_masked
    mask_data["P_MASKED"] = n_masked / (cutout.shape[0] * cutout.shape[1])

    return masked_cutout, mask_data


def generate_mask(cutout, nsigma=1., gauss_width=2.0, npixels=5):
    """ Generates a given mask based on the input parameters


    Args:
        cutout:     input cutout (2D-Array)
        nsigma:     Number of sigma levels above the measured background pixels are required to be
                    at to be flagged as belonging to an object.
        gauss_width: The gaussian width of the smoothing kernal for the mask.
        npixels:    The number of connected pixels required in order for something to be considered an object in the
                    segmentation map.

    Returns:
        The segment array as a 2D array
    """

    sigma = gauss_width * gaussian_fwhm_to_sigma
    kernel = Gaussian2DKernel(sigma).normalize()

    # Find threshold for cutout, and make segmentation map
    threshold = detect_threshold(cutout, nsigma=nsigma)
    segments = detect_sources(cutout, threshold, npixels=npixels, filter_kernel=kernel)

    # Attempt to de-blend. Return original segments upon failure.
    try:
        deb_segments = deblend_sources(cutout, segments, npixels=npixels, filter_kernel=kernel)
    except ImportError:
        print("Skimage not working!")
        deb_segments = segments
    except:
        # Don't do anything if it doesn't work
        deb_segments = segments

    segment_array = deb_segments.data

    return segment_array


def boolean_mask(mask, omit=None):
    """ Turns a given mask (photutils segment array) into a boolean array)

    Args:
        omit: If None, mask everything. Otherwise if an int or a list, do not mask those specified.
    """
    if omit is None:
        omit = []
    elif type(omit) == int:
        omit = [omit]

    bool_mask = ndarray(mask.shape, dtype="bool")

    bool_mask[:] = False
    bool_mask[mask > 0] = True
    for val in omit:
        bool_mask[mask == val] = False

    return bool_mask


def mask_cutouts(cutouts, config=None, method='standard', progress_bar=False):
    """ Mask a set of cutouts according to a certain method.

    Args:
        cutouts:
        config: User configuration file.
        method: Which method to use.
            standard : normal method. regular mask parameters. omits mask on central object
            no_central: regular mask parameters. masks central object
            background: background method: more severe mask parmeters. masks central object
        progress_bar: Use a TQDM progress bar.

    Returns:
        list of masked cutouts
    """
    masked_cutouts = []
    bg_means, bg_medians, bg_stds = [], [], []
    iterable = cutouts if not progress_bar else tqdm(cutouts)

    for cutout in iterable:
        try:
            if method == 'standard':
                masked, mask_data = mask_cutout(cutout, config=config, omit_centre=True)
            elif method == 'no_central':
                masked, mask_data = mask_cutout(cutout, config=config, omit_centre=False)
            elif method == 'background':
                masked, mask_data = mask_cutout(cutout, config=None,
                                                nsigma=0.5, gauss_width=2.0, npixels=5, omit_centre=False)
            else:
                continue

            masked_cutouts.append(masked)
            bg_means.append(mask_data["BG_MEAN"])
            bg_medians.append(mask_data["BG_MEDIAN"])
            bg_stds.append(mask_data["BG_STD"])
        except AttributeError:
            print("Cutout might be a NoneType")

    return masked_cutouts, [bg_means, bg_medians, bg_stds]



""" Surface brightness profile methods.

This module contains all methods pertaining directly to surface brightness
profiles measured using isophote fitting.

"""


from numpy import trapz, copy, isnan, isinf, log10, arange, min, max, gradient, median, std
from scipy.interpolate import interp1d
from scipy.signal import savgol_filter
from scipy.stats import norm


def get_half_light_radius(x, y):
    """ Obtains the index of the ellipse whose semimajor axis is at the half-light radius.

    Args:
        x (arr): Array of radii
        y (arr): Array of intensities

    Returns:
        best_index: the index containing the array value closest to the half-light radius
    """
    # Estimates the half-light radius for a given profile.

    half_light = trapz(y, x) / 2
    best_sec = 0     # Light contained in given section
    best_index = 0
    for index in range(0, len(x)):
        sec = trapz(y[:index], x[:index])
        if abs(sec - half_light) < abs(best_sec - half_light):
            best_sec, best_index = sec, index
    return best_index


def sb_to_i(sb, m_0=27, a_pix=1):
    """ Convert a given surface brightness to an intensity.

    Args:
        sb (float): Surface brightness
        m_0 (float): Zero-point magnitude
        a_pix (float): Area of a pixel (in arc seconds squared)

    Returns:
        Intensity value for the given surface brightness.
    """
    exponential = (sb - m_0) / -2.5
    return a_pix * (10 ** exponential)


def adjust_profile(x, y, shift=0.00, clean=False, log_x=False):
    """ Adjusts a profile to a common plotting format (surface brightness and optional shift vs regular pixels).

    Args:
        x (arr): Array of radii
        y (arr): Array of intensities
        shift (float): Introduce a shift to the intensity values (sometimes helpful if many negative numbers present)
        clean (bool): If true, clean nan and inf values from array (optional)
        log_x (bool): If True, log the radial intensities.
    Returns:
        x_new, y_new : The adjusted radial and intensity arrays
    """

    x_new, y_new = copy(x), copy(y)
    if log_x:
        x_new = log10(x_new)

    y_new = -2.5 * log10((y + shift) / (0.168 ** 2)) + 27

    if clean:
        x_clean, y_clean = [], []
        for n in range(0, len(x_new)):
            if isnan(y_new[n]) or isinf(y_new[n]):
                continue
            else:
                x_clean.append(x_new[n])
                y_clean.append(y_new[n])
        profile_adj = interp1d(x_clean, y_clean, bounds_error=False, fill_value=max(y_clean))

        return x_new, profile_adj(x_new)

    else:
        return x_new, y_new


def overlap(sma, low_1, high_1, low_2, high_2, threshold=1):
    """ Tests for non_overlap between array sets 1 and 2

    Args:
        sma (arr): Test parameters, preferably a numpy arange object
        low_1 (scipy.interpolate.interp1d): interp object for the lower bounds of set 1
        high_1 (scipy.interpolate.interp1d): interp object for the higher bounds of set 1
        low_2 (scipy.interpolate.interp1d): interp object for the lower bounds of set 2
        high_2 (scipy.interpolate.interp1d): interp object for the higher bounds of set 2
        threshold (int): How many consecutive times you see non-overlap between the two profiles.

    Returns:
        The index where the sufficient overlap is broken.
    """
    count, i = 0, 0
    for i in sma:
        overlapping = not ((high_2(i) < low_1(i)) or (low_2(i) > high_1(i)))
        if overlapping:
            count = 0
        if count == threshold:
            break
        count += 1

    return i


def smooth(arr, window=9, poly=1, iterations=2):
    """ Smooth an array with some number of Savitzky-Golay filters

    Args:
        arr (arr): Array to be be smoothed
        window (int): Filter window length
        poly (int): Order of polynomail
        iterations (int): The number of times to smooth the array

    Returns:
        Smoothed array of same length as arr

    """
    arr = copy(arr)
    for i in range(iterations):
        arr = savgol_filter(arr, window_length=window, polyorder=poly)

    return arr


def profile_quality(profile, bg_mean, bg_std, alpha=0.05, seeing=0, intens_cut=0.01):
    """ Determines the quality of a given profile compared to measured background levels.

    Args:
        profile (isotable): The input profile (still in the photutils isotable format).
        bg_mean (float): The background mean
        bg_std (float): The background standard deviation
        alpha (float): The confidence interval to define a "good" measurement
        seeing (float): The inner seeing distance within which we will not consider measurements to count.
        intens_cut (float): A minimum intensity beyond which we will stop counting measurements, inputted
                            as a fraction of the central profile intensity.

    Returns:
        good (int): The number of good measurements in the profile.
        bad (int): The number of bad measurements in the profile.
    """
    # Retrieve the background distribution as a norm dist. centered on bg_mean with std=bg_std
    bg_distribution = norm(loc=bg_mean, scale=bg_std)

    good, bad, samples = 0, 0, len(profile)
    sma_table, intens_table = profile["sma"], profile["intens"]

    max_intens = max(intens_table)
    intens_cut = max_intens * intens_cut

    # Iterate over all intensity measurements
    # If the cumulative dist. function at that point is greater than 1-alpha, then we
    # have a good measurement. Otherwise, we have a bad measurement.

    for i in range(samples):
        # Don't count measurements that are made within the seeing HWHM
        if sma_table[i] < seeing:
            continue

        intens = intens_table[i]
        if intens <= intens_cut:
            break
        if bg_distribution.sf(intens) < alpha:
            good += 1
        else:
            bad += 1

    return good, bad


def profile_asymptote(x, y, step=0.5, smooth=True, min_cut=10, threshold=1e-4):
    """ Determine the profile asymptotic value from a profile's x and y parameters.

    This method smooths out the valid data with a Savitzky-Golay filter then determines statistics on the
    values where the gradient is sufficiently low (below the threshold).

    Args:
        x (arr): Array of radii
        y (arr): Array of intensities
        step (float): When resampling the arrays, this is the step between samples.
        smooth (bool): If True, smoothing will be applied. (Recommended).
        min_cut (int): The number of measurements at the start of the profile to ignore. This is to remove any issues
                       with low gradients at the start of the profile.
        threshold (int): The gradient threshold below which measurements will be counted as sufficiently asymptotic.

    Returns:
        med (float): The median asymptote value
        std (float): The standard deviation of the asymptote measurements
        range (tuple): The range of locations (radial distances) where measurements are taken

    """
    interp = interp1d(x, y, fill_value=min(y), bounds_error=False)
    x_new = arange(min(x), max(x) + step, step)
    y_new = interp(x_new)

    # Just remove any nan values (we'll assume they their removal will not cause major problems)
    x_new, y_new = x_new[min_cut:], y_new[min_cut:]
    x_new, y_new = x_new[~isnan(y_new)], y_new[~isnan(y_new)]

    if smooth:
        # Twice smooth the data
        smoothed = savgol_filter(y_new, window_length=9, polyorder=2)
        smoothed = savgol_filter(smoothed, window_length=9, polyorder=2)
        y_new = smoothed

    grad = abs(gradient(y_new, x_new))
    y_good = y_new[grad < threshold]
    x_good = x_new[grad < threshold]

    if len(y_good) == 0:
        print("No values available to calculate gradient.")
        return None

    # Return the asymptote, the spread of values in asymptotic numbers, and the range of "good" measurements
    return median(y_good), std(y_good), (min(x_good), max(x_good))



""" Background estimation and subtraction module.

This module contains all methods available to estimate the backgrounds of cutouts,
and to subtract these backgrounds from cutouts or profiles as a means to correct them.

"""

import numpy as np
from astropy.table import Table
from astropy.stats import sigma_clipped_stats, SigmaClip
from astropy.convolution import convolve, Tophat2DKernel

from photutils.aperture import CircularAnnulus, EllipticalAnnulus
from photutils import Background2D, MedianBackground, BkgZoomInterpolator, make_source_mask


def background_2D(img, mask, box_size, interp=None, filter_size=1,
                  exclude_percentile=90):
    """ Run photutils background with SigmaClip and MedianBackground

    :param img: The 2D input image
    :type img: array_like (np.ndarray)
    :param mask: The 2D image mask
    :type mask: array_like (np.ndarray)
    :param box_size: The size of the box used in the 2D background. It should be larger than
        the largest objects in your image but still sufficiently small to capture large-scale
        structure.
    :type box_size: int, optional
    :param filter_size: The window size of the median filter being applied to the background image. A
        higher filter size will result in more smoothing to the background.
    :type filter_size: int, optional
    :param exclude_percentile: If the percentage of masked pixels in a box is above the exclude percentile,
        it is not included in determining the 2D background.
    :type exclude_percentile: float, optional
    """
    if interp is None:
        interp = BkgZoomInterpolator()
    return Background2D(img, box_size,
                        sigma_clip=SigmaClip(sigma=3.),
                        filter_size=filter_size,
                        bkg_estimator=MedianBackground(),
                        exclude_percentile=exclude_percentile,
                        mask=mask,
                        interpolator=interp)


class SourceMask:
    def __init__(self, img, nsigma=3., npixels=10, mask=None):
        """ Helper for making & dilating a source mask.
             See Photutils docs for make_source_mask.

            :param img: The image that is to be masked
            :type img: array_like (np.ndarray)
            :param nsigma: The sigma detection threshold for the source mask, defaults to 3
            :type nsigma: float, optional
            :param npixels: The number of required pixels for a detection, defaults to 10
            :type npixels: int, optional
            :param mask: An already-suppled mask for objects in the image.
            :type array_like (np.ndarray), optional
        """
        self.img = img
        self.nsigma = nsigma
        self.npixels = npixels
        if mask is None:
            self.mask = np.zeros(self.img.shape, dtype=np.bool)
        else:
            self.mask = mask

    def single(self, filter_fwhm=3., tophat_size=5., mask=None):
        """ Mask on a single scale """
        if mask is None:
            image = self.img
        else:
            image = self.img * (1 - mask)
        mask = make_source_mask(image, nsigma=self.nsigma,
                                npixels=self.npixels,
                                dilate_size=1, filter_fwhm=filter_fwhm)
        return dilate_mask(mask, tophat_size)

    def multiple(self, filter_fwhm=[3.], tophat_size=[3.], mask=None):
        """ Mask repeatedly on different scales """
        if mask is None:
            self.mask = np.zeros(self.img.shape, dtype=np.bool)
        for fwhm, tophat in zip(filter_fwhm, tophat_size):
            smask = self.single(filter_fwhm=fwhm, tophat_size=tophat)
            self.mask = self.mask | smask  # Or the masks at each iteration

        return self.mask

    def dilated(self, tophat_size=[3.], mask=None):
        """Mask using simple dilation"""
        if mask is None:
            self.mask = self.single()
        for tophat in tophat_size:
            smask = dilate_mask(self.mask, tophat)
            self.mask = self.mask | smask  # Or the masks at each iteration

        return self.mask


def dilate_mask(mask, tophat_size):
    """ Dilate a mask with a tophat kernel. """
    area = np.pi * tophat_size ** 2.
    kernel = Tophat2DKernel(tophat_size)
    dilated_mask = convolve(mask, kernel) >= 1. / area
    return dilated_mask


def estimate_background(cutout, config, model_params=None):
    """
    Estimate the background for a cutout using some of the various available methods based on what is set in the
    configuration file.

    OPTIONS:
        ellipse: uses elliptical annuli to estimate the cutout
        circle: uses a circular annuli
        sigclip (DEFAULT): measures the background using sigma-clipping.

    :param cutout: The input cutout
    :type cutout: array_like (np.ndarray)
    :param config: The configuration file (generated with galprime.config)
    :type config: dict
    :param model_params: Input model params for the ellipse method
    :type model_params: dict, optional
    """
    if config["BG_PARAMS"] == "ellipse" and model_params is not None:
        bg_mean, bg_median, bg_std = estimate_bg_elliptical_annulus(cutout,
                                                                    ellipticity=model_params["ELLIP"],
                                                                    r_50=model_params["R50"],
                                                                    pa=model_params["PA"],
                                                                    width=50,
                                                                    factor=20)
    elif config["BG_PARAMS"] == 'circle':
        bg_mean, bg_median, bg_std = estimate_bg_annulus(cutout, dynamic=True, annulus_width=50)
    else:
        bg_mean, bg_median, bg_std = estimate_background_sigclip(cutout)

    return bg_mean, bg_median, bg_std


def estimate_background_sigclip(cutout, nsigma=2, npixels=3, dilate_size=7):
    """ Estimate the background mean, median, and standard deviation of a cutout using sigma-clipped-stats """

    bg_mask = make_source_mask(cutout, nsigma=nsigma, npixels=npixels, dilate_size=dilate_size)

    bg_mean, bg_median, bg_std = sigma_clipped_stats(cutout, sigma=3.0, mask=bg_mask)

    return bg_mean, bg_median, bg_std


def estimate_bg_annulus(cutout, annulus_radius=50, annulus_width=10, dynamic=True, **kwargs):
    """
    Measure the background of a cutout using a circular annulus

    :param annulus_radius: Radius of the inner component of the annulus, defaults to 50 pix
    :type annulus_radius: int, optional
    :param annulus_width: Width of the annulus, defaults to 10 pix
    :type annulus_width: int, optional
    :param dynamic: Automatically sets the size of the annulus to be as large as possible
    :type dynamic: bool, optional
    """

    args = {"nsigma": 2, "npixels": 5, "dilate_size": 11, "sigclip_iters": 5}
    for arg in kwargs:
        args[arg] = kwargs[arg]

    if dynamic:
        annulus_radius = (cutout.shape[0] / 2) - annulus_width
    # First generate a background source mask for the input cutout
    bg_mask = make_source_mask(cutout,
                               nsigma=args["nsigma"],
                               npixels=args["npixels"],
                               dilate_size=args["dilate_size"],
                               sigclip_iters=args["sigclip_iters"])

    # Generate the annulus
    annulus = CircularAnnulus([cutout.shape[0] / 2, cutout.shape[1] / 2],
                              r_in=annulus_radius,
                              r_out=annulus_radius + annulus_width)

    # Generate a mask from the annulus and ensure it can be properly added to the cutout
    annulus_mask = annulus.to_mask(method='center')
    annulus_mask = annulus_mask.to_image(shape=cutout.shape)
    annulus_mask[annulus_mask == 0] = np.nan

    # Get the background pixels and remove any masked pixels
    bg_pixels = cutout * annulus_mask
    bg_pixels[bg_mask] = np.nan

    # Run our basic statistics and return them
    mean, median, std = np.nanmean(bg_pixels), np.nanmedian(bg_pixels), np.nanstd(bg_pixels)
    return mean, median, std


def estimate_bg_elliptical_annulus(cutout, ellipticity=0, r_50=50, pa=0, width=20, factor=10, return_mask=False):
    """
    Measure the background of a cutout using an elliptical annulus.

    :param cutout: The input cutout
    :type cutout: array_like (np.ndarray)
    :param ellipticity: The ellipticity of the elliptical annulus (e = 1 - b/a), defaults to 0
    :type ellipticity: float, optional
    :param r_50: The half light radius in pixels
    :type r_50: float, optional
    :param pa: the position angle in degrees. 0 is along the x-axis, defaults to 0
    :type pa: float, optional
    :param width: The width of the elliptical annulus in pixels, defaults to 20
    :type width: float, optional
    :param factor: The multiples of r_50 in which to increase the size of the elliptical annulus, defaults to 10
    :type factor: float, optional
    :param return_mask: If true, will return the mask alongsize the background statistics, defaults to False.
    :type return_mask: bool, optional
    :return: If return_mask is False, returns (mean, median, std)
        If return_mask is True, returns (mean, median, std, mask)
    """
    a_in = r_50 * factor
    b_in = a_in * (1 - ellipticity)

    a_out, b_out = a_in + width, b_in + width

    bg_mask = make_source_mask(cutout, nsigma=2, npixels=2, dilate_size=11)

    annulus = EllipticalAnnulus([cutout.shape[0] / 2, cutout.shape[1] / 2],
                                a_in=a_in, a_out=a_out,
                                b_in=b_in, b_out=b_out,
                                theta=pa)

    annulus_mask = annulus.to_mask(method='center')
    annulus_mask = annulus_mask.to_image(shape=cutout.shape)
    annulus_mask[annulus_mask == 0] = np.nan

    bg_pixels = cutout * annulus_mask
    bg_pixels[bg_mask] = np.nan

    mean, median, std = np.nanmean(bg_pixels), np.nanmedian(bg_pixels), np.nanstd(bg_pixels)

    if return_mask:
        return mean, median, std, bg_pixels
    else:
        return mean, median, std


def estimate_background_set(cutouts):
    """
    Estimates the background values for a set of cutouts.
    :param cutouts: A list of array_like cutouts
    :type cutouts: array_like
    :return: The 3 arrays of background means, medians, and standard deviations
    :rtype: array_like
    """
    bg_means, bg_medians, bg_stds = [], [], []

    for cutout in cutouts:
        bg_mean, bg_median, bg_std = estimate_background_sigclip(cutout)

        bg_means.append(bg_mean)
        bg_medians.append(bg_median)
        bg_stds.append(bg_std)

    return bg_means, bg_medians, bg_stds


def subtract_backgrounds(profile_set, background_array):
    """
    Generate an array of tables identical to the input except the respective backgrounds
    are subtracted from the intensity array for each table.

    :param profile_set: Set of profiles (in the photutiuls isolist format)
    :type profile_set: list(isolist.to_table())
    :param background_array: Array of background values to subtract from each profile table.
    :type background_array: array_like
    :return: List of profiles of length len(profile_set)
    """
    bg_subtracted_tables = []

    for i in range(0, len(profile_set)):
        this_table = profile_set[i]
        isotable_localsub = Table()

        for col in this_table.colnames:
            isotable_localsub[col] = np.copy(this_table[col])
        isotable_localsub["intens"] = ((isotable_localsub["intens"]) - background_array[i])

        bg_subtracted_tables.append(isotable_localsub)

    return bg_subtracted_tables



import galprime
import galprime.plotting as plotter
from astropy.table import Table
from astropy.io import fits

import time
import traceback

from pebble import ProcessPool
from concurrent.futures import TimeoutError

from numpy import round, array
from numpy.random import choice, randint, seed


def pipeline(config, max_bins=None, mag_table=None,
             provided_bgs=None, provided_psfs=None,
             progress_bar=False, multiprocess_level='obj'):
    """
    Runs the entire simulation pipeline assuming certain data exists.
    :param config: Values from properly loaded configuration file.
    :param max_bins: The number of bins to process (useful if running tests).
    :param mag_table: Optional array of magnitudes.
    :param mag_table_keys Keys for optional mag table binning.
    :param provided_bgs: A set of provided background cutouts [OPTIONAL].
    :param provided_psfs: A set of provided PSFs related to the provided backgrounds [OPTIONAL].
    :param progress_bar: Have a TQDM progress bar.
    :param multiprocess_level: Where in the simulations to divide into cores
        'obj'  - Divide at the object level, where each core handles a single object in each bin.
        'bin'  - Divide at the bin level, so each core is responsible for a single bin
        'none' - Do not do any multiprocessing at all (SLOW).

        The object level is less memory intensive, but bins are processed one by one instead of simultaneously.

    EXAMPLE USAGE:

    config_values = tbridge.load_config_file("path/to/config/file.tbridge")
    tbridge.pipeline(config_values, max_bins=10)

    """

    binned_objects = galprime.bin_catalog(config)
    max_bins = len(binned_objects) if max_bins is None else max_bins

    verbose = config["VERBOSE"]
    if verbose:
        print(max_bins, "bins to process.")

    if config["SAME_BGS"] and provided_bgs is None:
        provided_bgs, provided_psfs = galprime.get_backgrounds(config, n=int(config["N_BGS"]),
                                                              return_psfs=True, return_bg_info=False)
        if verbose:
            print(len(provided_bgs), "obtained with associated PSFs.")

    for b in binned_objects[:max_bins]:

        if mag_table is not None:
            separate_mags = galprime.bin_mag_catalog(mag_table, b,
                                                    mag_table_keys=[config["MASS_KEY"], config["Z_KEY"]],
                                                    bin_keys=["MASSES", "REDSHIFTS"])[config["MAG_KEY"]]
        else:
            separate_mags = None

        _process_bin(b, config, separate_mags=separate_mags,
                     provided_bgs=provided_bgs, provided_psfs=provided_psfs,
                     progress_bar=progress_bar, multiprocess=True)

    galprime.config_to_file(config, filename=config["OUT_DIR"] + "tbridge_config.txt")


def _process_bin(b, config_values, separate_mags=None, provided_bgs=None, provided_psfs=None,
                        progress_bar=False, multiprocess=False, profiles_per_row=3):
    """
        Process a single bin of galaxies. (Tuned for pipeline usage but can be used on an individual basis.
        :param b: Bin to obtain full profiles from.
        :param config_values: Values from properly loaded configuration file.
        :param separate_mags: Optional array of magnitudes.
        :param provided_bgs: Array of provided backgrounds
        :param progress_bar: Use a TQDM progress bar (note with multithreading this might get funky).
        :param multiprocess: Run in multiprocessing mode.
            This means both the model creation AND the
        :return:
        """
    t_start = time.time()
    verbose = config_values["VERBOSE"]

    # Load in bin information, and prepare all necessary structural parameters.
    keys, columns = b.return_columns()
    mags, r50s, ns, ellips = galprime.pdf_resample(galprime.structural_parameters(keys, columns), resample_size=1000)
    if separate_mags is not None:
        mags = galprime.pdf_resample(separate_mags, resample_size=len(r50s))[0]

    # Simulate the model rows, using multiprocessing to speed things up#
    if verbose:
        print("Processing", config_values["N_MODELS"], "models for: ", b.bin_params)

    # Prepare containers for simulations
    job_list = []
    full_profile_list, bg_infolist, cutout_infolist, bg_2D_profiles = [[] for i in range(profiles_per_row)], [], [], []
    masked_cutouts, unmasked_cutouts = [], []
    param_results = Table(data=None, names=["MAG", "R50", "I_R50", "N", "ELLIP", "PA"])

    # Get the original models
    models, model_parameters = galprime.simulate_sersic_models(mags, r50s, ns, ellips,
                                                              config_values, n_models=config_values["N_MODELS"])

    # Run multithreaded simulation code
    with ProcessPool(max_workers=config_values["CORES"]) as pool:
        for i in range(len(models)):
            job_list.append(pool.schedule(_process_model,
                                          args=(models[i], config_values, model_parameters[i],
                                                provided_bgs, provided_psfs),
                                          timeout=config_values["ALARM_TIME"]))
    # Collect the results
    for i in range(len(job_list)):
        try:
            result = job_list[i].result()
            profiles = result["PROFILES"]
            if len(profiles) != profiles_per_row:
                continue
            # If we got enough profiles, append everything to the appropriate arrays
            for i in range(len(full_profile_list)):
                full_profile_list[i].append(profiles[i])

            parameters = result["INFO"]
            row = [parameters[key] for key in param_results.colnames]
            param_results.add_row(row)

            bg_infolist.append(result["BG_DATA"])
            cutout_infolist.append(result["CUTOUT_DATA"])

            # masked_cutouts.append(result["MASKED_CUTOUT"])
            # unmasked_cutouts.append(result["UNMASKED_CUTOUT"])
            bg_2D_profiles.append(result["BGSUB_PROFILE"])

        except Exception as error:
            if config_values["TEST_VERBOSE"]:
                print(error.args, i, traceback.print_exc())
            else:
                print(error.args, i)

    bg_info, cutout_info = [[], [], []], [[], [], []]

    for i in range(0, len(bg_infolist)):
        bg_info[0].append(bg_infolist[i]["BG_MEAN"])
        bg_info[1].append(bg_infolist[i]["BG_MEDIAN"])
        bg_info[2].append(bg_infolist[i]["BG_STD"])

    for i in range(0, len(cutout_infolist)):
        cutout_info[0].append(cutout_infolist[i]["BG_MEAN"])
        cutout_info[1].append(cutout_infolist[i]["BG_MEDIAN"])
        cutout_info[2].append(cutout_infolist[i]["BG_STD"])

    # Subtract the median values from the bgadded profiles
    if config_values["BG_PARAMS"] == "2D":
        print("Using 2D profiles")
        bg_sub_profiles = bg_2D_profiles
    else:
        bg_sub_profiles = galprime.subtract_backgrounds(full_profile_list[2], bg_info[1])

    full_profile_list.append(bg_sub_profiles)

    if verbose:
        print("Finished extraction. Saving info for", len(full_profile_list[0]),
              "successful extractions out of", len(models), ".")

    # Save the profiles to the required places
    galprime.save_profiles(full_profile_list,
                          bin_info=b.bin_params,
                          out_dir=config_values["OUT_DIR"],
                          keys=["bare", "noisy", "bgadded", "bgsub"],
                          bg_info=bg_info, cutout_info=cutout_info,
                          structural_params=param_results)

    if verbose:
        print("Finished", b.bin_params, "-- Time Taken:", round((time.time() - t_start) / 60, 2), "minutes.")
        print()


def _process_model(sersic_model, config, model_params, provided_bgs=None, provided_psfs=None):
    """
    Run processing for a single model.

    :param sersic_model: The input model to process
    :param config: Configuration values loaded from config file
    :param provided_bgs: OPTIONAL --- set of provided backgrounds
    """
    # First make the input models

    if provided_bgs is None:
        background, psf, bg_data = galprime.get_background(config)
        convolved_model = galprime.convolve_models(sersic_model, psf=psf)
        bg_added_model = convolved_model + background

    else:
        if config["PROVIDED_BG_TYPE"] == "Cutouts":
            # We will now randomly select a background and psf, and use that as our background
            seed()
            bg_index = randint(0, len(provided_bgs),)
            convolved_model = galprime.convolve_models(sersic_model, psf=provided_psfs[bg_index])
            bg_added_model = convolved_model + provided_bgs[bg_index]

        else:
            seed()
            cutout, row_data = galprime.cutout_from_table(provided_bgs)

            with fits.open(config["PSF_FILENAME"]) as psfs:
                psf = galprime.get_closest_psf(psfs, row_data["RA"], row_data["DEC"])
            convolved_model = galprime.convolve_models(sersic_model, config_values=config, psf=psf)
            bg_added_model = convolved_model + cutout

    # Estimate the background using the sigclip method before we do anything to it with 2D Background subtraction
    bg_mean, bg_median, bg_std = galprime.estimate_background_sigclip(bg_added_model)
    bg_data = {"BG_MEAN": bg_mean, "BG_MEDIAN": bg_median, "BG_STD": bg_std}

    # If we're doing 2D background subtraction we'll need to estimate and then subtract the background BEFORE
    # profile extraction.
    bgsub_model = None
    if config["BG_PARAMS"] == "2D":
        sm = galprime.SourceMask(bg_added_model)
        mask = sm.multiple(filter_fwhm=[1, 3, 5], tophat_size=[4, 2, 1])
        bg = galprime.background_2D(bg_added_model, mask,
                                   box_size=config["BOX_SIZE"], filter_size=config["FILTER_SIZE"])
        bgsub_model = bg_added_model - bg.background

    # Now we will make the cutout smaller for extraction purposes
    convolved_model = galprime.central_cutout(convolved_model, width=config["EXTRACTION_SIZE"])
    bg_added_model = galprime.central_cutout(bg_added_model, width=config["EXTRACTION_SIZE"])
    bgsub_model = galprime.central_cutout(bgsub_model, width=config["EXTRACTION_SIZE"])

    # Generate the noise model and mask the background-added model
    noisy_model = galprime.add_to_noise(convolved_model)
    masked_model, mask_data = galprime.mask_cutout(bg_added_model, config=config)
    model_row = [convolved_model, noisy_model, masked_model]

    # Then extract everything we can
    profile_extractions = []
    for i in range(0, len(model_row)):
        try:
            extraction = galprime.isophote_fitting(model_row[i], config=config)
            profile_extractions.append(extraction.to_table())
        except Exception as error:
            print(error.args)
            continue

    # We'll extract the BG-subtracted profile is necessary
    bgsub_profile = None
    if config["BG_PARAMS"] == "2D":
        try:
            bgsub_profile = galprime.isophote_fitting(galprime.mask_cutout(bgsub_model, config=config)[0],
                                                     config=config).to_table()
        except Exception as error:
            print(error.args)

    # put the results into a dictionary format and return everything
    return {"PROFILES": profile_extractions,
            "BGSUB_PROFILE": bgsub_profile,
            "MASK_DATA": mask_data,
            "BG_DATA": bg_data,
            "CUTOUT_DATA": {"BG_MEAN": bg_mean, "BG_MEDIAN": bg_median, "BG_STD": bg_std},
            # "UNMASKED_CUTOUT": bg_added_model,
            # "MASKED_CUTOUT": masked_model,
            "INFO": model_params}



import galprime

from numpy import mgrid, copy
from skimage.filters import gaussian
from numpy import sqrt, sort


def cutout_gradient(cutout, depth_value=5):
    """
    Estimate the gradient of a first-smoothed cutout according to
    the following formula:

        G(x,y) = ( I(x,y) - I_0 )^2 / d
            where d is the distance from (x,y) to the cutout centre.
        The returned gradient is the Nth largest gradient value, where N is
        defined by the user as depth_value.

    Args:
        cutout: The cutout to measure
        depth_value: Nth largest gradient value to return (default is 5).

    Returns:
        The gradient of the submitted cutout.
    """
    x, y = mgrid[:cutout.shape[0], :cutout.shape[1]]
    cutout_centre = (cutout.shape[0] / 2, cutout.shape[1] / 2)
    c_x, c_y = int(cutout_centre[0]), int(cutout_centre[1])

    cutout = gaussian(cutout, sigma=5)

    dists = sqrt((cutout_centre[0] - x) ** 2 + (cutout_centre[1] - y) ** 2)

    central_value = cutout[c_x][c_y]

    grads = (cutout - central_value) ** 2 / dists
    grads = sort(grads, axis=None)

    return grads[len(grads) - 1 - depth_value]


def cutout_gradients(cutouts, mask_cutouts=False):
    """
    Return the gradients for a set of cutouts. (Convenience function).

    Args:
        cutouts: List of cutouts.
        mask_cutouts: Set to True if the input cutouts need to have light from
            nearby objects masked first.

    Returns:
        list: the gradients for the supplied cutouts.
    """

    if mask_cutouts:
        cutouts, bg_info = galprime.mask_cutouts(cutouts, method='no_central')

    gradients = []
    for cutout in cutouts:
        gradients.append(cutout_gradient(cutout))
    return gradients



from numpy import arange, nan, inf, sort, median, floor, max, nanmedian, transpose, asarray, std
from numpy.random import choice, randint
from scipy.interpolate import interp1d

from astropy.io import fits
from astropy.table import Table

import galprime
import os
import shutil
from pathlib import Path


def as_interpolations(profile_list, fill_value_type='min', x_key="sma", y_key="intens"):

    interps = []

    for prof in profile_list:
        sma, intens = prof[x_key], prof[y_key]
        try:
            interp = interp1d(sma, intens, bounds_error=False, fill_value=0)
            interps.append(interp)
        except ValueError:
            continue
    return interps


def bin_max(profile_list, key="sma"):
    max_val = -999

    for prof in profile_list:
        arr_max = max(prof[key])
        max_val = arr_max if arr_max > max_val else max_val

    return max_val


def save_medians(median_data, bootstrap_data=None, output_filename="medians.fits"):
    median_sma, median_interp = median_data
    median_intens = median_interp(median_sma)

    out_hdulist = fits.HDUList()

    t = Table([median_sma, median_intens], names=["SMA", "INTENS"])
    out_hdulist.append(fits.BinTableHDU(t))
    if bootstrap_data is not None:
        b_sma, b_1sig_l, b_1sig_u, b_2sig_l, b_2sig_u, b_3sig_l, b_3sig_u, b_5sig_l, b_5sig_u = bootstrap_data
        # Append Lower Bootstrap Value
        t = Table([b_sma, b_1sig_l(b_sma), b_2sig_l(b_sma), b_3sig_l(b_sma), b_5sig_l(b_sma)],
                  names=["SMA", "INTENS_1SIG", "INTENS_2SIG", "INTENS_3SIG", "INTENS_5SIG"])
        out_hdulist.append(fits.BinTableHDU(t))
        # Append Upper Bootstrap Value
        t = Table([b_sma, b_1sig_u(b_sma), b_2sig_u(b_sma), b_3sig_u(b_sma), b_5sig_u(b_sma)],
                  names=["SMA", "INTENS_1SIG", "INTENS_2SIG", "INTENS_3SIG", "INTENS_5SIG"])
        out_hdulist.append(fits.BinTableHDU(t))

    out_hdulist.writeto(output_filename, overwrite=True)


def index_format(in_dir, x_bins, y_bins, out_dir="dir_copy/", indices=(1, 2), method='duplicate'):
    """
    Duplicate a directory but with the files in index format.
    :param in_dir: Input directory of median objects
    :param out_dir: Only needed if running in duplicate mode. Duplicates files in new directory
    :param x_bins: x-axis bin values (usually grabbed from config file)
    :param y_bins: y-axis bin values
    :param indices: When splitting the file, the indices correspond to x and y parameter locations.
        For example, if your filename is: bin_9.2-9.6_0.3-0.5_0.0-0.5_bare.fits
        and you want the x parameter to be 9.2-9.6 and the y parameter to be 0.3-0.5, you would
        set:
            indices=[1,2]
    :param method: duplicate directory or rename files
        'duplicate' : generate a new directory
        'rename' : Rename the files in input directory
    :return:
    """
    subdirs = os.listdir(in_dir)

    if method == 'duplicate' and not os.path.isdir(out_dir):
        os.mkdir(out_dir)

    for subdir in subdirs:

        if os.path.isfile(in_dir + subdir):
            continue

        subdir += "/"
        files = os.listdir(in_dir + subdir)

        if method == 'duplicate' and not os.path.isdir(out_dir + subdir):
            os.mkdir(out_dir + subdir)

        for f in files:
            splits = f.split("_")
            x_split = splits[indices[0]].split("-")
            x_val = (float(x_split[0]) + float(x_split[1])) / 2
            y_split = splits[indices[1]].split("-")
            y_val = (float(y_split[0]) + float(y_split[1])) / 2

            x_index, y_index = galprime.bin_index(x_val, x_bins), galprime.bin_index(y_val, y_bins)

            # print(f, "|", x_val, x_index, y_val, y_index)
            new_filename = str(x_index) + "_" + str(y_index) + "_" + f.split("_")[-1]
            if method == 'duplicate':
                shutil.copyfile(in_dir + subdir + f, out_dir + subdir + new_filename)
            elif method == 'rename':
                os.rename(in_dir + subdir + f, in_dir + subdir + new_filename)


def load_median_info(filename):
    """
    Load a complete set of median info (TBriDGE output) in as a dict object.
    :param filename:
    :return:
    """
    median_data = {}

    hdul = fits.open(filename)
    med, l, u = Table.read(hdul[1]), Table.read(hdul[2]), Table.read(hdul[3])

    median_data["MED_SMA"] = med["SMA"]
    median_data["MED_INTENS"] = med["INTENS"]

    median_data["MED_ADJ"] = galprime.adjust_profile(med["SMA"], med["INTENS"])

    median_data["L_1SIG"] = (l["SMA"], l["INTENS_1SIG"])
    median_data["U_1SIG"] = (u["SMA"], u["INTENS_1SIG"])
    median_data["L_1SIG_ADJ"] = galprime.adjust_profile(l["SMA"], l["INTENS_1SIG"])
    median_data["U_1SIG_ADJ"] = galprime.adjust_profile(u["SMA"], u["INTENS_1SIG"])

    median_data["L_2SIG"] = (l["SMA"], l["INTENS_2SIG"])
    median_data["U_2SIG"] = (u["SMA"], u["INTENS_2SIG"])
    median_data["L_2SIG_ADJ"] = galprime.adjust_profile(l["SMA"], l["INTENS_2SIG"])
    median_data["U_2SIG_ADJ"] = galprime.adjust_profile(u["SMA"], u["INTENS_2SIG"])

    median_data["L_3SIG"] = (l["SMA"], l["INTENS_3SIG"])
    median_data["U_3SIG"] = (u["SMA"], u["INTENS_3SIG"])
    median_data["L_3SIG_ADJ"] = galprime.adjust_profile(l["SMA"], l["INTENS_3SIG"])
    median_data["U_3SIG_ADJ"] = galprime.adjust_profile(u["SMA"], u["INTENS_3SIG"])

    median_data["L_5SIG"] = (l["SMA"], l["INTENS_5SIG"])
    median_data["U_5SIG"] = (u["SMA"], u["INTENS_5SIG"])
    median_data["L_5SIG_ADJ"] = galprime.adjust_profile(l["SMA"], l["INTENS_5SIG"])
    median_data["U_5SIG_ADJ"] = galprime.adjust_profile(u["SMA"], u["INTENS_5SIG"])

    return median_data


def bin_max(profiles, key="sma"):
    bin_max = -999
    for prof in profiles:
        try:
            prof_max = prof[key][-1]
            if prof_max > bin_max:
                bin_max = prof_max
        except Exception as error:
            continue
    return bin_max


def normalize_bin(profiles, bin_max, step=0.5, key="sma"):
    new_sma = arange(0, bin_max + step, step)
    normalized_profiles = [prof(new_sma) for prof in profiles]

    return new_sma, asarray(normalized_profiles)


def normalized_median(normalized_profiles, handle_nans=True):
    if handle_nans:
        median_prof = nanmedian(normalized_profiles, axis=0)
    else:
        median_prof = median(normalized_profiles, axis=0)
    # print(median_prof.shape)
    return median_prof


def normalized_bootstrap(normalized_profiles, iterations=1001, gaussian=True, handle_nans=True):
    true_median = normalized_median(normalized_profiles)

    medians = []
    bin_size = len(normalized_profiles)
    for i in range(iterations):
        pop = normalized_profiles[randint(0, bin_size, size=bin_size)]
        medians.append(normalized_median(pop, handle_nans=handle_nans))

    medians = asarray(medians)

    # If we can assume a Gaussian distribution, this part is easy
    if gaussian:
        slices = std(transpose(medians), axis=1)

        lower_1sig = true_median - slices
        upper_1sig = true_median + slices

        lower_2sig = true_median - 2 * slices
        upper_2sig = true_median + 2 * slices

        lower_3sig = true_median - 3 * slices
        upper_3sig = true_median + 3 * slices

        lower_5sig = true_median - 5 * slices
        upper_5sig = true_median + 5 * slices

        return {"L_1SIG": lower_1sig, "L_2SIG": lower_2sig, "L_3SIG": lower_3sig, "L_5SIG": lower_5sig,
                "U_1SIG": upper_1sig, "U_2SIG": upper_2sig, "U_3SIG": upper_3sig, "U_5SIG": upper_5sig}

    else:
        slices = sort(transpose(medians))

        lower_index_1sig, upper_index_1sig = int(iterations * 0.159), int(iterations * 0.841)
        lower_index_2sig, upper_index_2sig = int(iterations * 0.023), int(iterations * 0.977)
        lower_index_3sig, upper_index_3sig = int(iterations * 0.002), int(iterations * 0.998)

        lower_1sig = [median_slice[lower_index_1sig] for median_slice in slices]
        upper_1sig = [median_slice[upper_index_1sig] for median_slice in slices]

        lower_2sig = [median_slice[lower_index_2sig] for median_slice in slices]
        upper_2sig = [median_slice[upper_index_2sig] for median_slice in slices]

        lower_3sig = [median_slice[lower_index_3sig] for median_slice in slices]
        upper_3sig = [median_slice[upper_index_3sig] for median_slice in slices]

        lower_5sig = [median_slice[0] for median_slice in slices]
        upper_5sig = [median_slice[len(median_slice) - 1] for median_slice in slices]

        return {"L_1SIG": lower_1sig, "L_2SIG": lower_2sig, "L_3SIG": lower_3sig, "L_5SIG": lower_5sig,
                "U_1SIG": upper_1sig, "U_2SIG": upper_2sig, "U_3SIG": upper_3sig, "U_5SIG": upper_5sig}


def median_pipeline(full_filename, output_filename=None, step=0.5):
    profiles = galprime.load_profile_set(full_filename)

    bin_max_value = bin_max(profiles)
    median_sma, normalized_profiles = normalize_bin(galprime.as_interpolations(profiles), bin_max_value,
                                                    step=step)

    median_profile = normalized_median(normalized_profiles)
    b_dict = normalized_bootstrap(normalized_profiles)

    if output_filename is not None:
        out_hdulist = fits.HDUList()

        med_table = Table([median_sma, median_profile], names=["SMA", "INTENS"])

        b_lower_table = Table([median_sma, b_dict["L_1SIG"],
                               b_dict["L_2SIG"], b_dict["L_3SIG"], b_dict["L_5SIG"]],
                              names=["SMA", "INTENS_1SIG", "INTENS_2SIG", "INTENS_3SIG", "INTENS_5SIG"])

        b_upper_table = Table([median_sma, b_dict["U_1SIG"], b_dict["U_2SIG"],
                               b_dict["U_3SIG"], b_dict["U_5SIG"]],
                              names=["SMA", "INTENS_1SIG", "INTENS_2SIG", "INTENS_3SIG", "INTENS_5SIG"])

        out_hdulist.append(fits.BinTableHDU(med_table))
        out_hdulist.append(fits.BinTableHDU(b_lower_table))
        out_hdulist.append(fits.BinTableHDU(b_upper_table))

        out_hdulist.writeto(output_filename, overwrite=True)

        return b_dict.update({"MEDIAN": med_table, "SMA": median_sma})





""" Binning methods

This module contains all necessary methods for binning catalogues according to various parameters.

"""

from astropy.table import Table
from numpy import array, copy, append, round, reshape


class Bin:
    """ Class for a single bin of information
        :param objects: The catalog rows that belong to the given bin
        :type bin_param_names: array_like
        :param object_column_names: The column names when the objects are sorted column-wise instead of row-wise.
        :type bin_param_names: array_like
        :param bin_params (arr): The values that define the bounds of the bin.
        :type bin_param_names: array_like
        :param bin_param_names (arr): Array of parameter names that the bin was made with.
        :type bin_param_names: array_like
    """

    def __init__(self, objects=None, object_column_names=None, bin_params=None, bin_param_names=None):
        """ Constructor method
        """
        if bin_params is None:
            bin_params = []
        if bin_param_names is None:
            bin_param_names = []

        self.objects = objects
        self.object_column_names = object_column_names
        self.bin_params = bin_params
        self.bin_param_names = bin_param_names

    def rebin(self, param_index, low_bounds, bin_width, number_threshold=4):
        """ Returns a list of Bin objects after re-binning based on a certain value and bin parameters.

        Bins specified by the lower bounds and bin width.

        :param param_index: Which index in the objects contains the desired value.
        :type param_index: int
        :param low_bounds: Lower bounds on the bins.
        :type low bounds: array_like
        :param bin_width: The width of the bins.
        :type bin_width: float
        param number_threshold: The required number of objects a bin needs to have to be saved to the list.

        Returns:
             List of re-binned Bin objects.
        """
        new_bins = []
        top_level_params = self.bin_params
        top_level_param_names = self.bin_param_names

        for low in low_bounds:
            this_bin = []

            for obj in self.objects:
                if low < obj[param_index] < low + bin_width:
                    this_bin.append(obj)
            if len(this_bin) <= number_threshold:
                continue

            new_bin = Bin(objects=this_bin, object_column_names=self.object_column_names,
                          bin_params=append(copy(top_level_params), (round(low, 2), round(low + bin_width, 2))),
                          bin_param_names=append(copy(top_level_param_names), self.key_from_index(param_index)))
            new_bins.append(new_bin)

        return new_bins

    def return_columns(self):
        """
        Returns a dictionary of columns (turns the object list back into a column format)
        """
        keys = copy(self.object_column_names)
        columns = []
        for i in range(0, len(keys)):
            columns.append([])
            for j in range(0, len(self.objects)):
                columns[i].append(self.objects[j][i])

        return list(keys), columns

    def index_from_key(self, key=""):
        """ Return the index of the column name key specified."""
        return self.object_column_names.index(key)

    def key_from_index(self, index=0):
        """ Returns the column name at a given index. """
        return self.object_column_names[index]

    def values_in_bin(self, values):
        """ Returns a boolean whether of not a set of values is within the bin.

        :param values: A list of values.
        :type values: arr_like
        """
        reshaped_values = reshape(self.bin_params, newshape=(int(len(self.bin_params) / 2), 2))
        good_count = 0
        for i in range(0, len(values)):
            if reshaped_values[i][0] < values[i] < reshaped_values[i][1]:
                good_count += 1

        if good_count == len(values):
            return True
        else:
            return False

    def return_param_dict(self):
        """
        Returns a dict object where the keys are the column names that built the parameter set,
        and the values are the corresponding parameters.
        """
        param_dict = {}
        for i in range(0, len(self.bin_param_names)):
            param_dict[self.bin_param_names[i]] = (self.bin_params[2 * i], self.bin_params[2 * i + 1])
        return param_dict


def select_bin(bin_list, values):
    """ Select a bin from a list of bins based on an array of values.

    Args:
        bin_list (arr): List of Bin objects
        values (arr): Array of parameters.
            Not that the parameters need to be in the same order that the bins were generated from.

    Returns:
        Bin corresponding to the appropriate parameters. If no bin exists for those parameters, returns None.

    """
    for b in bin_list:
        if b.values_in_bin(values):
            return b
    return None


def generate_objects(columns):
    """ Reorganizes the table columns into an object-format
    Args:
        columns: A list of table columns

    Returns:
        List of objects.
    """
    objects = []
    for i in range(0, len(columns[0])):
        this_obj = []
        for j in range(0, len(columns)):
            this_obj.append(columns[j][i])
        objects.append(this_obj)
    return objects


def bin_catalog(config):
    """ Bin the inputted catalog.

    This is a convenience function to get a list of bins based on redshift, star-formation, and mass.

    :param config: Values from input config file or user inputs.
    :type config: dict
    :return: List of bins that have been binned according to the config settings.
    """

    # Read HST-ZEST catalog into memory
    catalog = Table.read(config["CATALOG"], format="fits")

    # todo eventually we should change MASS_KEY, SFPROB_KEY, and Z_KEY into more abstract forms
    mags = array(catalog[config["MAG_KEY"]])
    r50s = array(catalog[config["R50_KEY"]])
    ns = array(catalog[config["N_KEY"]])
    ellips = array(catalog[config["ELLIP_KEY"]])
    masses = array(catalog[config["MASS_KEY"]])
    sfprobs = array(catalog[config["SFPROB_KEY"]])
    redshifts = array(catalog[config["Z_KEY"]])

    mass_bins, mass_step = config["MASS_BINS"], config["MASS_STEP"]
    redshift_bins, redshift_step = config["REDSHIFT_BINS"], config["REDSHIFT_STEP"]
    sfprob_bins, sfprob_step = config["SFPROB_BINS"], config["SFPROB_STEP"]

    catalog_objects = generate_objects((mags, r50s, ns, ellips, masses, redshifts, sfprobs))
    column_names = ["MAGS", "R50S", "NS", "ELLIPS", "MASSES", "REDSHIFTS", "SFPROBS"]
    init_bin = Bin(objects=catalog_objects, object_column_names=column_names, bin_params=[])

    mass_bins = init_bin.rebin(4, mass_bins, mass_step)
    mz_bins, full_bins = [], []
    for cat_bin in mass_bins:
        mz_bins.extend(cat_bin.rebin(5, redshift_bins, redshift_step))
    for cat_bin in mz_bins:
        full_bins.extend(cat_bin.rebin(6, sfprob_bins, sfprob_step))

    return full_bins


def bin_mag_catalog(mag_table, b, mag_table_keys=None, bin_keys=None):
    """ Bin an external catalogue based on a bin's provided parameters.

    :param mag_table: The magnitude table to rebin
    :type mag_table: astropy.table.Table
    :param b: The Bin object
    :type b: galprime.Bin
    :param mag_table_keys: The magnitude table keys to use for binning.
    :type mag_table_keys: arr_like
    :param bin_keys: The associated bin keys (must be the same length AND order as mag_table_keys).
    :type bin_keys: arr_like
    :return: Table of magnitudes binned by the appropriate parameters.
    """
    if bin_keys is None:
        bin_keys = ["MASSES", "REDSHIFTS"]
    if mag_table_keys is None:
        mag_table_keys = ["MASS", "Z"]

    # Gather the necessary information from the provided bin.
    bin_dict = b.return_param_dict()

    # Iterate over all keys and only return the rows that are within the bin limits.
    for i in range(len(bin_keys)):
        low, high = bin_dict[bin_keys[i]]
        mag_table = mag_table[mag_table[mag_table_keys[i]] >= low]
        mag_table = mag_table[mag_table[mag_table_keys[i]] <= high]

    return mag_table


def bin_index(val, bins):
    """ Get bin index for a given value and a set of bin parameters

    :param val: The value corresponding to a bin index
    :param bins: The list of bins to search through

    """
    for index in range(0, len(bins) - 1):
        if bins[index] < val < bins[index + 1]:
            return index
    return len(bins) - 1



from astropy.io import fits
from astropy.modeling.models import Gaussian2D, Sersic2D
from astropy.modeling import Fittable2DModel, Parameter
from astropy.stats import sigma_clipped_stats

from numpy import exp, isnan, mgrid, ceil, pi, cosh, cos, sin, sqrt, std, ndarray
from numpy.random import choice, randint, uniform

from photutils.datasets import make_noise_image

from scipy.signal import convolve2d
from scipy.special import gamma, gammainc, gammaincinv, kn
from scipy.optimize import newton
from scipy.stats import gaussian_kde

import galprime


def b(n, estimate=False):
    """ Get the b_n normalization constant for the sersic profile. From Graham and Driver."""
    if estimate:
        return 2 * n - (1 / 3) + (4 / (405 * n)) + (46 / (25515 * (n ** 2)))
    else:
        return gammaincinv(2 * n, 0.5)


def core_sersic_b(n, r_b, r_e):
    """
    Get the scale parameter b using the Newton-Raphson root finder.
    :param n: Sersic index
    :param r_b: Break radius
    :param r_e: Effective radius
    :return:
    """
    # Start by getting an initial guess at b (using the regular Sersic estimation)
    b_guess = 2 * n - (1 / 3)

    # Define the combination of gamma functions that makes up the relation
    # We want to find the zeroes of this.
    def evaluate(b_in):
        comp1 = gamma(2 * n)
        comp2 = gammainc(2 * n, b_in * ((r_b / r_e) ** (1 / n)))
        comp3 = 2 * gammainc(2 * n, b_in)

        return comp1 + comp2 - comp3

    return newton(evaluate, x0=b_guess)


def i_at_r50(mag, n=2, r_50=2, m_0=27):
    """ Get the intensity at the half-light radius """
    b_n = b(n)
    l_tot = 10 ** ((mag - m_0) / -2.5) * (b_n ** (2 * n))
    denom = (r_50 ** 2) * 2 * pi * n * exp(b_n) * gamma(2 * n)
    i_e = l_tot / denom

    return i_e


def pdf_resample(columns, resample_size=1000):
    """
    Makes a multidimensional probability distribution function (PDF) for galaxy structural parameters.
    :param columns: columns to resample
    :param resample_size: number of objects to resample
    :return:
    """

    obj_resample = gaussian_kde(columns).resample(resample_size)

    return obj_resample


def structural_parameters(keys, columns):
    """
    Gets the structural parameters in a specific order (if catalog is generated by the
        method in tbridge.binning)
    """
    mags = columns[keys.index("MAGS")]
    r50s = columns[keys.index("R50S")]
    ns = columns[keys.index("NS")]
    ellips = columns[keys.index("ELLIPS")]

    return mags, r50s, ns, ellips


def simulate_sersic_models(mags, r50s, ns, ellips, config_values, n_models=10):
    """
    Simulates Sersic models WITHOUT CONVOLUTION
    :param mags: object magnitudes
    :param r50s: object half-light radii (in arcseconds)
    :param ns: object Sersic indices
    :param ellips: object ellipticities
    :param config_values: configuration file information
    :param n_models: number of models to generate
    :return:
    """

    cutout_size, arc_conv = config_values["SIZE"], config_values["ARC_CONV"]

    # Prep mgrid
    x, y = mgrid[:cutout_size, :cutout_size]

    # Resample parameters and get a better distribution.
    mags, r50s, ns, ellips = galprime.pdf_resample((mags, r50s, ns, ellips), 1000)

    sersic_models, param_dicts = [], []
    for i in range(n_models):
        fails, clean = 0, False
        this_r50, this_n, this_ellip, this_mag, this_r50_pix, i_r50 = None, None, None, None, None, None
        while not clean:
            if fails > 10000:
                break
            obj_choice = randint(0, len(mags),)
            this_mag = mags[obj_choice]
            this_r50 = r50s[obj_choice]  # This is in arcseconds
            this_r50_pix = this_r50 / arc_conv  # This is in pixels
            this_n = ns[obj_choice]
            this_ellip = ellips[obj_choice]

            # Get the intensity at the half-light radius.
            i_r50 = i_at_r50(mag=this_mag, n=this_n, r_50=this_r50_pix)
            clean = not (i_r50 > 100) and not (isnan(i_r50)) and (this_n > 0.65) and (0 < this_ellip < 1) \
                    and (this_r50_pix > 0)
            if not clean:
                fails += 1
            if fails > 10000:
                continue

        pa = uniform(0, 2 * pi)
        struct_param_dict = {"I_R50": i_r50,
                             "R50": this_r50_pix,
                             "N": this_n,
                             "ELLIP": this_ellip,
                             "PA": pa,
                             "MAG": this_mag}

        sersic_model = Sersic2D(amplitude=i_r50, r_eff=this_r50_pix, n=this_n,
                                ellip=this_ellip, theta=pa,
                                x_0=cutout_size / 2, y_0=cutout_size / 2)
        z = sersic_model(x, y)
        sersic_models.append(z)
        param_dicts.append(struct_param_dict)

    return sersic_models, param_dicts


def add_to_background(model, config, convolve=True, return_bg_info=False, threshold=1e-4):
    """
    Add a model to a given background
    """
    image_dir, psf_filename = config["IMAGE_DIRECTORY"], config["PSF_FILENAME"]
    image_filenames = galprime.get_image_filenames(image_dir)
    bg_infotable = {"IMAGES": [], "RAS": [], "DECS": [], "XS": [], "YS": []}

    with fits.open(psf_filename) as psfs:
        model_width = model.shape[0]
        model_halfwidth = ceil(model_width / 2)

        image_filename = choice(image_filenames)

        fail_counter = 0

        while 1:
            fail_counter += 1
            if fail_counter > 50:
                return (None, None) if convolve else None

            image = galprime.select_image(image_filename)
            image_wcs = galprime.get_wcs(image_filename)

            c_x = randint(model_halfwidth + 1, image.shape[0] - model_halfwidth - 1)
            c_y = randint(model_halfwidth + 1, image.shape[1] - model_halfwidth - 1)
            x_min, x_max = int(c_x - model_halfwidth), int(c_x + model_halfwidth)
            y_min, y_max = int(c_y - model_halfwidth), int(c_y + model_halfwidth)

            image_cutout = image[x_min: x_max - 1, y_min: y_max - 1]

            # Check if the background std is less than a given threshold
            if threshold is not None:
                bg_mean, bg_median, bg_std = sigma_clipped_stats(image_cutout, sigma=3.)
                if bg_std < threshold:
                    continue
                else:
                    break

        ra, dec = image_wcs.wcs_pix2world(c_x, c_y, 0)
        psf = galprime.get_closest_psf(psfs, ra, dec).data

        if return_bg_info:
            bg_infotable["IMAGES"].append(image_filename)
            bg_infotable["RAS"].append(ra)
            bg_infotable["DECS"].append(dec)
            bg_infotable["XS"].append(c_x)
            bg_infotable["YS"].append(c_y)


        if convolve:
            convolved = convolve2d(model, psf, mode='same')
            bg_added = convolved + image_cutout

            return bg_added, convolved
        else:
            bg_added = model + image_cutout
            return bg_added, None


def get_background(config, threshold=1e-4):
    """
        Get a valid background and background info
        """
    image_dir, psf_filename = config["IMAGE_DIRECTORY"], config["PSF_FILENAME"]
    image_filenames = galprime.get_image_filenames(image_dir)
    bg_infotable = {}

    with fits.open(psf_filename) as psfs:
        model_width = config["SIZE"]
        model_halfwidth = ceil(model_width / 2)

        image_filename = choice(image_filenames)

        fail_counter = 0

        while 1:
            fail_counter += 1
            if fail_counter > 50:
                return None

            image = galprime.select_image(image_filename)
            image_wcs = galprime.get_wcs(image_filename)

            c_x = randint(model_halfwidth + 1, image.shape[0] - model_halfwidth - 1)
            c_y = randint(model_halfwidth + 1, image.shape[1] - model_halfwidth - 1)
            x_min, x_max = int(c_x - model_halfwidth), int(c_x + model_halfwidth)
            y_min, y_max = int(c_y - model_halfwidth), int(c_y + model_halfwidth)

            image_cutout = image[x_min: x_max - 1, y_min: y_max - 1]

            # Check if the background std is less than a given threshold
            if threshold is not None:
                bg_mean, bg_median, bg_std = galprime.estimate_background_sigclip(image_cutout)
                if bg_std < threshold:
                    continue
                else:
                    break
        ra, dec = image_wcs.wcs_pix2world(c_x, c_y, 0)
        psf = galprime.get_closest_psf(psfs, ra, dec).data

        bg_infotable["IMAGES"] = image_filename
        bg_infotable["RAS"] = ra
        bg_infotable["DECS"] = dec
        bg_infotable["XS"] = c_x
        bg_infotable["YS"] = c_y
        bg_infotable["BG_MEAN"] = bg_mean
        bg_infotable["BG_MEDIAN"] = bg_median
        bg_infotable["BG_STD"] = bg_std

    return image_cutout, psf, bg_infotable


def convolve_models(models, config_values=None, psf=None):
    """
    Convolve a set of models.
    :param models: List of cutouts to convolve.
    :param config_values:
    :param psf:
    :return:
    """

    if config_values is not None:
        psf_filename = config_values["PSF_FILENAME"]
        psfs = fits.open(psf_filename)
        convolved_models = []

        if type(models) == ndarray:
            index = randint(0, len(psfs))
            return convolve2d(models, psfs[index].data, mode='same')
        else:
            for model in models:
                index = randint(0, len(psfs))
                convolved_models.append(convolve2d(model, psfs[index].data, mode='same'))
        psfs.close()
    elif psf is not None:
        if type(models) == ndarray:
            return convolve2d(models, psf, mode='same')
        else:
            convolved_models = []
            for model in models:
                convolved_models.append(convolve2d(model, psf, mode='same'))
    else:
        return None

    return convolved_models


def add_to_provided_backgrounds(models, backgrounds):
    bgadded_models, bg_datatables = [], []

    if type(models) == ndarray:
        index = randint(0, len(backgrounds))

        return models + backgrounds[index]
    else:
        for model in models:
            index = randint(0, len(backgrounds))
            bgadded_models.append(model + backgrounds[index])

    return bgadded_models


def add_to_noise(models, bg_mean=0., bg_std=0.025):
    """
    Returns the models added to Gaussian noise with user-defined mean and standard deviation
    :param models: The models to add noise to.
    :param bg_mean: Mean of the Gaussian noise.
    :param bg_std: Standard deviation of the Gaussian noise.
    :return:
    """
    noisy_images = []

    if type(models) == ndarray:
        noise_image = make_noise_image((models.shape[0], models.shape[1]), mean=bg_mean, stddev=bg_std)
        return models + noise_image
    else:
        for model in models:
            noise_image = make_noise_image((model.shape[0], model.shape[1]), mean=bg_mean, stddev=bg_std)
            noisy_images.append(model + noise_image)
        return noisy_images


def simulate_bg_gaussians(n_bgs, n_models, width=251, noise_mean=0.001, noise_std=0.01,
                          amplitude=(0, 3), stddev=(2, 10), min_dist=0):
    x_0 = width /2
    x, y = mgrid[:width, :width]
    backgrounds = []
    for i in range(0, n_bgs):
        # Generate base image with Gaussian noise
        random_bg = make_noise_image((width, width), distribution='gaussian', mean=noise_mean, stddev=noise_std)

        # Generate a random assortment of Gaussian models with the supplied User parameters
        for j in range(0, n_models):
            x_mean, y_mean, dist = uniform(0, width), uniform(0, width), 0
            while dist <= min_dist:
                x_mean, y_mean = uniform(0, width), uniform(0, width)
                dist = sqrt((x_mean - x_0) ** 2 + (y_mean - x_0) ** 2)

            model = Gaussian2D(amplitude=uniform(amplitude[0], amplitude[1]),
                               x_mean=x_mean,
                               y_mean=y_mean,
                               x_stddev=uniform(stddev[0], stddev[1]),
                               y_stddev=uniform(stddev[0], stddev[1]),
                               theta=uniform(0, 2 * pi))
            z = model(x, y)
            random_bg += z
        backgrounds.append(random_bg)
    return backgrounds


class EdgeOnDisk(Fittable2DModel):
    """
    Two-dimensional Edge-On Disk model.

    Parameters
    ----------
    amplitude : float
        Brightness at galaxy centre
    scale_x : float
        Scale length along the semi-major axis
    scale_y : float
        Scale length along the semi-minor axis
    x_0 : float, optional
        x position of the center
    y_0 : float, optional
        y position of the center
    theta: float, optional
        Position angle in radians, counterclockwise from the
        positive x-axis.
    """
    amplitude = Parameter(default=1)
    scale_x = Parameter(default=1)
    scale_y = Parameter(default=0.5)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    theta = Parameter(default=0)

    @classmethod
    def evaluate(cls, x, y, amplitude, scale_x, scale_y, x_0, y_0, theta):
        """Exaluate model on a 2D x-y grid."""

        x_maj = abs((x - x_0) * cos(theta) + (y - y_0) * sin(theta))
        x_min = -(x - x_0) * sin(theta) + (y - y_0) * cos(theta)

        return amplitude * (x_maj / scale_x) * kn(1, x_maj / scale_x) / (cosh(x_min / scale_y) ** 2)


class Core_Sersic(Fittable2DModel):
    """
    Two-dimensional Edge-On Disk model.

    Parameters
    ----------
    r_e : float
        Effective radius of the galaxy.
    r_b : float
        Break Radius (Where the model switches from one regime to the other).
    I_b : float
        Intensity at the break radius
    alpha : float
        Defines the "sharpness" of the model transitions
    gamma : float
        Power law slope
    n     : float
        Sersic index (see info on Sersic profiles if needed)
    x_0 : float, optional
        x position of the center
    y_0 : float, optional
        y position of the center
    theta: float, optional
        Position angle in radians, counterclockwise from the positive x-axis.
    ellip: float, optional
        Ellipticity of the model (default is 0 : circular)

    """

    r_e = Parameter(default=5)
    r_b = Parameter(default=1)
    I_b = Parameter(default=1)
    n = Parameter(default=1)
    alpha = Parameter(default=1)
    gamma = Parameter(default=1)

    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)
    theta = Parameter(default=0)
    ellip = Parameter(default=0)

    @classmethod
    def evaluate(cls, x, y, r_e, r_b, I_b, n, alpha, gamma, x_0, y_0, theta, ellip):
        """Two dimensional Core-Sersic profile function."""

        bn = core_sersic_b(n, r_b, r_e)

        def core_sersic_i_prime():
            return I_b * (2 ** (- gamma / alpha)) * exp(bn * (2 ** (1 / (alpha * n))) * (r_b / r_e) ** (1 / n))

        i_prime = core_sersic_i_prime()

        cos_theta, sin_theta = cos(theta), sin(theta)
        x_maj = (x - x_0) * cos_theta + (y - y_0) * sin_theta
        x_min = -(x - x_0) * sin_theta + (y - y_0) * cos_theta
        z = sqrt(x_maj ** 2 + (x_min / (1-ellip)) ** 2)

        comp_1 = i_prime * ((1 + ((r_b / z) ** alpha)) ** (gamma / alpha))
        comp_2 = -bn * (((z ** alpha) + (r_b ** alpha)) / (r_e ** alpha)) ** (1 / (n * alpha))

        return comp_1 * exp(comp_2)


class ObjectGenError(Exception):
    """ Error raised if fails exceed 10000"""

    def __init__(self, expression="", message="Fails exceeded the limit"):
        self.expression = expression
        self.message = message



"""
This module contains all general-purpose functions for loading and saving data. This includes pulling data out of
FITS files, saving TBriDGE simulation outputs, and more.
"""

import galprime
import os
import math
from pathlib import Path

from astropy.io import fits
from astropy.wcs import wcs
from astropy.table import Table
from astropy.nddata import Cutout2D, NoOverlapError
from astropy.stats import sigma_clipped_stats

from numpy import sqrt, str, save, load, ceil, zeros, isnan, floor, ceil, sort, min, mean
from numpy.random import choice, randint, uniform


def get_closest_psf(psfs, obj_ra, obj_dec):
    """ Get the closest psf for a given object's RA and DEC"""

    def dist(ra1, ra2, dec1, dec2):
        return sqrt((ra1 - ra2) ** 2 + (dec1 - dec2) ** 2)

    shortest_dist = 999999
    closest_psf = None

    # Cycle through psfs and look for the best one
    for psf in psfs:
        head = psf.header
        psf_ra, psf_dec = head["RA"], head["DEC"]
        local_dist = dist(obj_ra, psf_ra, obj_dec, psf_dec)
        if local_dist < shortest_dist:
            shortest_dist = local_dist
            closest_psf = psf

    return closest_psf


def get_wcs(fits_filename):
    """
        Finds and returns the WCS for an image. If Primary Header WCS no good, searches each index until a good one
        is found. If none found, raises a ValueError
    """
    # Try just opening the initial header
    wcs_init = wcs.WCS(fits_filename)
    ra, dec = wcs_init.axis_type_names
    if ra.upper() == "RA" and dec.upper() == "DEC":
        return wcs_init

    else:
        hdu_list = fits.open(fits_filename)
        for n in hdu_list:
            try:
                wcs_slice = wcs.WCS(n.header)
                ra, dec = wcs_slice.axis_type_names
                if ra.upper() == "RA" and dec.upper() == "DEC":
                    return wcs_slice
            except:
                continue
        hdu_list.close()

    raise ValueError


def get_arc_conv(w: wcs.WCS):
    """ Gets pixels to arc-seconds conversion scale (Number of arcseconds per pixel) """
    pix_x, pix_y = 1, 1
    ra_1, dec_1 = w.wcs_pix2world(pix_x, pix_y, 0)
    ra_2, dec_2 = w.wcs_pix2world(pix_x + 1, pix_y + 1, 0)
    diff_1 = abs(ra_1 - ra_2) * 3600
    diff_2 = abs(dec_1 - dec_2) * 3600
    return (diff_1 + diff_2) / 2


def get_angular_size_dist(z, H0=71, WM=0.27, WV=None):
    """
    Return the angular size distance in Megaparsecs.
    (Stripped down version of Cosmocalc by Ned Wright and Tom Aldcroft (aldcroft@head.cfa.harvard.edu))
    """
    try:
        c = 299792.458  # velocity of light in km/sec

        if z > 100:
            z /= 299792.458  # Values over 100 are in km/s

        if WV is None:
            WV = 1.0 - WM - 0.4165 / (H0 * H0)  # Omega(vacuum) or lambda
        age = 0.0  # age of Universe in units of 1/H0

        h = H0 / 100.
        WR = 4.165E-5 / (h * h)  # includes 3 massless neutrino species, T0 = 2.72528
        WK = 1 - WM - WR - WV
        az = 1.0 / (1 + 1.0 * z)
        n = 1000  # number of points in integrals
        for i in range(n):
            a = az * (i + 0.5) / n
            adot = math.sqrt(WK + (WM / a) + (WR / (a * a)) + (WV * a * a))
            age += 1. / adot

        DCMR = 0.0

        # do integral over a=1/(1+z) from az to 1 in n steps, midpoint rule
        for i in range(n):
            a = az + (1 - az) * (i + 0.5) / n
            adot = math.sqrt(WK + (WM / a) + (WR / (a * a)) + (WV * a * a))
            DCMR = DCMR + 1. / (a * adot)

        DCMR = (1. - az) * DCMR / n

        # tangential comoving distance
        ratio = 1.0
        x = math.sqrt(abs(WK)) * DCMR
        if x > 0.1:
            if WK > 0:
                ratio = 0.5 * (math.exp(x) - math.exp(-x)) / x
            else:
                ratio = math.math.sin(x) / x
        else:
            y = x * x
            if WK < 0:
                y = -y
            ratio = 1. + y / 6. + y * y / 120.
        DCMT = ratio * DCMR
        DA = az * DCMT
        Mpc = lambda x: c / H0 * x
        DA_Mpc = Mpc(DA)

        return DA_Mpc
    except:
        raise ValueError


def generate_cutout(image, position, img_wcs=None, size=91, world_coords=True):
    """
    Generates a cutout for a given image. Uses world coordinates by default, but can be configured to take
    in a position corresponding to the actual array indices.
    :param image: The input image.
    :type image: array_like
    :param position: The central position of the cutout.
        It is in (RA,DEC) if using world coordinates.
        Otherwise it is (pix_x, pix_y), and the user must set world_coords to True.
    :type position: arr_like
    :param img_wcs: The world coordinate system associated with the image. Needed if inputting RA,DEC
        ccordinates.
    :type img_wcs: astropy.wcs.WCS object, optional
    :param size: The size of the cutout, defaults to 91 pixels
    :type size: int, optional
    :param world_coords: Whether position corresponds to an RA,DEC or not. Defaults to True
    :type world_coords: bool, optional
    :return:
    """

    if size < 71:
        size = 71
    elif size % 2 == 0:
        size += 1

    if world_coords:
        coord = img_wcs.wcs_world2pix(position[0], position[1], 0, ra_dec_order=True)
        pix_x, pix_y = coord[0], coord[1]
    else:
        pix_x, pix_y = position[0], position[1]

    # Send an index error iff the position does not lie in the image.
    if 0 < pix_x < image.shape[0] and 0 < pix_y < image.shape[1]:
        try:
            cut = Cutout2D(image, (pix_x, pix_y), size)
            return cut.data
        except NoOverlapError:
            raise IndexError
    else:
        raise IndexError


def trim_objects(obj_catalog, ras, decs, mag_lim=None, ra_key="RA", dec_key="DEC", mag_key="i"):
    this_catalog = obj_catalog[obj_catalog[ra_key] > ras[0]]
    this_catalog = this_catalog[this_catalog[ra_key] < ras[1]]

    this_catalog = this_catalog[this_catalog[dec_key] > decs[0]]
    this_catalog = this_catalog[this_catalog[dec_key] < decs[1]]

    if mag_lim is not None:
        this_catalog = this_catalog[this_catalog[mag_key] > mag_lim[0]]
        this_catalog = this_catalog[this_catalog[mag_key] < mag_lim[1]]

    return this_catalog


def extraction_limits(image_wcs, image_shape, cutout_size):
    """ Get the minimum and maximum RA and DEC limits for a given cutout size and image
        Used when trimming a catalogue to only the usable objects.

        :param image_wcs:
        :type image_wcs: astropy.wcs.WCS object
        :param image_shape: The shape of the input image
        :type image_shape: tuple
        :param cutout_size: The width of the cutouts used in extraction
        :type cutout_size: int
    """
    half_width = ceil(cutout_size / 2)

    pix_min = image_wcs.wcs_pix2world(half_width, half_width, 0)
    pix_max = image_wcs.wcs_pix2world(image_shape[0] - half_width, image_shape[1] - half_width, 0)

    ras = sort((pix_min[0], pix_max[0]))
    decs = sort((pix_min[1], pix_max[1]))

    return ras, decs


def get_image_filenames(images_directory, image_band="i", check_band=False):
    """
    Retrieves a list of all available filenames for a given directory, and a given band.
    WARNING: Optimized for HSC filename format (ex: HSC-I_9813_4c3.fits).
    """
    image_filenames = []
    images = Path(images_directory).rglob('*.fits')
    for image in images:
        image = str(image)
        if check_band:
            image_no_path = image.split("/")[len(image.split("/")) - 1]
            filename_band = image_no_path.split("_")[0].split("-")[1].lower()
            if filename_band == image_band:
                image_filenames.append(image)
        else:
            image_filenames.append(image)
    return image_filenames


def select_image(filename):
    """ Opens an image-based FITS file and returns the first available image.

    Args:
        filename (str): The FITS filename and path.

    Returns:
        numpy.ndarray: The image obtained from the FITS file. If none found, will return None.

    """
    with fits.open(filename) as HDUList:
        image = None
        for i in range(0, len(HDUList)):
            try:
                image = HDUList[i].data
                if image.shape[0] > 0 and image.shape[1] > 0:
                    break
            except:
                continue
    return image


def generate_file_prefix(bin_params):
    """ Use the bin params to generate a file prefix."""
    prefix = "bin_"
    for j in range(0, len(bin_params)):
        if (j + 1) % 2 != 0:
            prefix += str(bin_params[j]) + "-"
        else:
            prefix += str(bin_params[j]) + "_"
    return prefix


def save_profiles(profile_list, bin_info, out_dir, keys, bg_info=None, cutout_info=None, structural_params=None):
    """ Saves a set of profiles into a properly formatted output directory, with proper filename format.

    Args:
        profile_list: The list of profiles (shape is m x n, where m is the number of different models for each
            object and n is the number of objects i.e. m rows of profiles from n objects.
        bin_info (arr): bin information for profile formatting (Obtained using the bin_params attribute from
            the tbridge.binning.Bin object).
        out_dir (str): the output directory to save the files to.
        keys (arr): the keys to generate subdirectory and file names with.
        bg_info (numpy.ndarray): Array of associated background info for the extracted profiles. Saves to its own
                    subdirectory.
        structural_params (astropy.table.Table): Astropy table of associated structural parameters
    """

    # Generate output structure
    subdirs = []
    if not os.path.isdir(out_dir):
        os.mkdir(out_dir)
    for key in keys:
        subdir = key + "_profiles/"
        if not os.path.isdir(out_dir + subdir):
            os.mkdir(out_dir + subdir)
        subdirs.append(subdir)

    filename_prefix = generate_file_prefix(bin_info)
    valid_colnames = ["sma", "intens", "intens_err", "ellipticity", "ellipticity_err", "pa", "pa_err"]

    # Save profiles as FITS HDULists in the proper directories
    for i in range(0, len(profile_list)):
        profiles = profile_list[i]
        out_filename = filename_prefix + keys[i] + ".fits"
        out_hdulist = fits.HDUList()
        for prof in profiles:
            try:
                out_hdulist.append(fits.BinTableHDU(Table([prof[col] for col in valid_colnames],
                                                          names=valid_colnames)))
            except Exception as e:
                print(e)
                continue

        out_hdulist.writeto(out_dir + subdirs[i] + out_filename, overwrite=True)

    # Save the additional available information
    if bg_info is not None:
        if not os.path.isdir(out_dir + "bg_info/"):
            os.mkdir(out_dir + "bg_info/")
        galprime.save_array(bg_info, out_dir + "bg_info/" + filename_prefix + "bgs.npy")
    if cutout_info is not None:
        if not os.path.isdir(out_dir + "cutout_info/"):
            os.mkdir(out_dir + "cutout_info/")
        galprime.save_array(cutout_info, out_dir + "cutout_info/" + filename_prefix + "bgs.npy")

    if structural_params is not None:
        if not os.path.isdir(out_dir + "params/"):
            os.mkdir(out_dir + "params/")
        structural_params.write(out_dir + "params/" + galprime.generate_file_prefix(bin_info) + "params.fits",
                                format="fits", overwrite=True)


def save_profile_set(profiles, out_filename="profiles.fits"):
    """ Save an array of profiles to a FITS file.

    Args:
        profiles: List of profile tables.
        out_filename: Filename to save FITS file to.

    """

    valid_colnames = ["sma", "intens", "intens_err", "ellipticity", "ellipticity_err", "pa", "pa_err"]
    out_hdulist = fits.HDUList()
    for prof in profiles:
        out_hdulist.append(fits.BinTableHDU(Table([prof[col] for col in valid_colnames],
                               names=valid_colnames)))

    out_hdulist.writeto(out_filename, overwrite=True)


def load_profile_set(filename):
    """ Load a set of profiles from a FITS file specified by the filename"""
    tables = []
    with fits.open(filename) as HDUList:
        for hdu in HDUList:
            try:
                tables.append(Table.read(hdu))
            except:
                continue
    return tables


def save_cutouts(cutouts, output_filename="cutouts.fits"):
    """ Save a set of cutouts to a fits HDUList object """
    out_hdulist = fits.HDUList()

    for cutout in cutouts:
        out_hdulist.append(fits.ImageHDU(data=cutout))

    out_hdulist.writeto(output_filename, overwrite=True)


def load_cutouts(filename):
    """ Load in a list of cutouts from the given filename (will try all hdus in the HDUList)"""
    cutouts = []
    HDUList = fits.open(filename)
    for n in HDUList:
        try:
            image = n.data
            if image.shape[0] > 0 and image.shape[1] > 0:
                cutouts.append(image)
        except:
            continue

    HDUList.close()

    return cutouts


def tables_from_file(filename):
    """
    Load a set of tables from a given HDUList.
    :param filename: filename to gather tables from.
    :return:
    """
    tables = []

    HDUList = fits.open(filename)
    for n in HDUList:
        try:
            t = Table.read(n)
            tables.append(t)
        except:
            continue
    HDUList.close()

    return tables


def params_from_filename(filename):
    """ Return the parameters inferred from the input filename (based on TBRIDGE formatting)."""
    filename_no_path = filename.split("/")[len(filename.split("/")) - 1]
    splits = filename_no_path.split("_")
    splits = splits[1:len(splits) - 1]
    params = []
    for split in splits:
        low_high = split.split("-")
        params.append((float(low_high[0]) + float(low_high[1])) / 2)
    return params


def save_array(arr, filename="arr.npy"):
    """ Save an array (npy binary format)"""
    save(filename, arr)


def load_array(filename):
    """ Load an array """
    try:
        return load(filename)
    except FileNotFoundError:
        print("File: ", filename, " not found. Returning NoneType")
        return None


def get_backgrounds(config, n=50, size=None, threshold=1e-4, return_psfs=True, return_bg_info=True):
    """ Retrieve a random set of backgrounds from the input image directory.
    Args:
        config_values: User provided configuration file.
        n: Number of backgrounds to generate.
        return_psfs (bool): Return the psfs associated with the backgrounds
        return_bg_info (bool): Return the background info.
    :return:
    """

    image_dir, psf_filename = config["IMAGE_DIRECTORY"], config["PSF_FILENAME"]
    size = config["SIZE"]

    with fits.open(psf_filename) as psfs:
        image_filenames = galprime.get_image_filenames(image_dir)

        psf_list, bg_added_models, image_list, ras, decs = [], [], [], [], []

        bg_infotable = {"IMAGES": [], "RAS": [], "DECS": [], "XS": [], "YS": []}

        for i in range(0, n):
            fail_count = 0
            while fail_count < 10:
                model_halfwidth = ceil(size / 2)

                image_filename = choice(image_filenames)

                image = galprime.select_image(image_filename)
                image_wcs = galprime.get_wcs(image_filename)

                c_x = randint(model_halfwidth + 1, image.shape[0] - model_halfwidth - 1)
                c_y = randint(model_halfwidth + 1, image.shape[1] - model_halfwidth - 1)
                x_min, x_max = int(c_x - model_halfwidth), int(c_x + model_halfwidth)
                y_min, y_max = int(c_y - model_halfwidth), int(c_y + model_halfwidth)

                image_cutout = image[x_min: x_max - 1, y_min: y_max - 1]
                ra, dec = image_wcs.wcs_pix2world(c_x, c_y, 0)

                if threshold is not None:
                    bg_mean, bg_median, bg_std = sigma_clipped_stats(image_cutout, sigma=3.)
                    if bg_std < threshold:
                        fail_count += 1
                        continue
                    else:
                        break

                if image is None:
                    fail_count += 1
                    continue

            if fail_count >= 10:
                continue

            if return_psfs:
                psf = galprime.get_closest_psf(psfs, ra, dec).data
                psf_list.append(psf)

            bg_added_models.append(image_cutout)
            bg_infotable["IMAGES"].append(image_filename)
            bg_infotable["RAS"].append(ra)
            bg_infotable["DECS"].append(dec)
            bg_infotable["XS"].append(c_x)
            bg_infotable["YS"].append(c_y)

    if return_psfs:
        if return_bg_info:
            return bg_added_models, psf_list, bg_infotable
        else:
            return bg_added_models, psf_list
    elif return_bg_info:
        return bg_added_models, bg_infotable
    else:
        return bg_added_models


def as_dir(directory):
    """ Add a forward slash if one is not at the end of a string. """
    if directory[-1] != '/':
        return directory + "/"
    else:
        return directory


def cutout_stitch(cutouts, masked_cutouts=None, output_filename=None):
    """ Generate a full stitch from a set of cutouts.

    Args:
        cutouts: Array of cutouts (ideally unmasked but it makes no difference). Need to be all the same size as the
            first cutout.
        masked_cutouts: The same cutouts but masked
        output_filename: Optional output to save as HDUList

    Returns:
        Returns either the stitch, or a tuple of the stich and the mask stitch if masked_cutouts is submitted..
    """

    i = int(sqrt(len(cutouts)))
    j = int(len(cutouts) / i)

    c_width, c_height = cutouts[0].shape[0], cutouts[1].shape[1]
    w, h = i * c_width, j * c_height

    canvas = zeros((w, h))

    if masked_cutouts is not None:
        masked_canvas = zeros((w, h))
    else:
        masked_canvas = None

    n = 0

    # Add cutouts to the canvas
    for x in range(i):
        for y in range(j):
            try:
                canvas[x * c_width: x * c_width + c_width,
                y * c_height: y * c_height + c_height] = cutouts[n]
            except ValueError:
                continue
            if masked_cutouts is not None:
                masked_cutout = masked_cutouts[n]
                mask = isnan(masked_cutout).astype(int)

                masked_canvas[x * c_width: x * c_width + c_width,
                y * c_height: y * c_height + c_height] = mask
            n += 1
            if n == len(cutouts):
                break

    if output_filename is not None:
        HDUList = fits.HDUList()
        HDUList.append(fits.ImageHDU(canvas))
        if masked_cutouts is not None:
            HDUList.append(fits.ImageHDU(masked_canvas))
        HDUList.writeto(output_filename, overwrite=True)

    if masked_cutouts is None:
        return canvas
    else:
        return canvas, masked_canvas


def central_cutout(arr, width=251):
    """
    Simple method to return a central cutout of a given width from an input array arr.
    """
    centre = [int(floor(arr.shape[0] / 2)), int(floor(arr.shape[1] / 2))]
    return Cutout2D(arr, centre, size=width).data


def get_background_table(directory, num_bgs=50, size=91, objects=None, min_dist=0,
                         attempts_per_img=5, output_filename=None, ra_key="RA", dec_key="DEC"):
    """ Generate a table where each row allows the user to reconstruct an image..

    Args:
        directory: Top level directory. This can have nested subdirectories.
        num_bgs: The number of backgrounds to denerate. Default is 50.
        size: Width of the cutout to generate.
        objects: Input table of objects to check for nearby objects.
        min_dist: The minimum allowable distance for an object, defined in the objects parameter,
            to be in order for a cutout to be added to the output table.
        attempts_per_img:
            The number

        output_filename: Optional output to save as HDUList

    Returns:
        Returns either the stitch, or a tuple of the stich and the mask stitch if masked_cutouts is submitted..
    """

    rows = []
    image_filenames = galprime.get_image_filenames(directory)

    attempts, max_attempts = 0, num_bgs * 100

    while len(rows) < num_bgs:
        if attempts >= max_attempts:
            break

        random_img = choice(image_filenames)

        image = galprime.select_image(random_img)
        image_wcs = galprime.get_wcs(random_img)

        for i in range(attempts_per_img):
            x = int(uniform(size + 1, image.shape[0] - size - 1))
            y = int(uniform(size + 1, image.shape[1] - size - 1))

            cutout = Cutout2D(image, (x, y), size=size, wcs=image_wcs)
            cutout_bbox = cutout.bbox_original[::-1]
            # print(cutout_bbox, [cutout_bbox[0][0], cutout_bbox[0][1], 0])

            ra, dec = image_wcs.wcs_pix2world(cutout_bbox[0], cutout_bbox[1], 0)
            ra, dec = sort(ra), sort(dec)

            ra_cent, dec_cent = image_wcs.wcs_pix2world(x, y, 0)
            # If we don't supply objects, just return the location. Otherwise, check for nearby objects
            if objects is None:
                rows.append([random_img, x, y, size, "N/A", "N/A", ra_cent, dec_cent])
                continue
            else:
                this_catalog = objects
                this_catalog = this_catalog[this_catalog[ra_key] > ra[0]]
                this_catalog = this_catalog[this_catalog[ra_key] < ra[1]]

                this_catalog = this_catalog[this_catalog[dec_key] > dec[0]]
                this_catalog = this_catalog[this_catalog[dec_key] < dec[1]]

                physical_position = image_wcs.wcs_world2pix(this_catalog[ra_key], this_catalog[dec_key], 0)
                phys_x, phys_y = physical_position[0], physical_position[1]

                try:
                    dists = sqrt((phys_x - x) ** 2 + (phys_y - y) ** 2)

                    if min(dists) < min_dist:
                        continue
                    else:
                        rows.append([random_img, x, y, size, min(dists), len(dists), ra_cent, dec_cent])
                except ValueError:
                    # Sometimes there are issues with the min function on the dists. Should probably look into this
                    # at some point.
                    continue

            # print(len(this_catalog), x, y, cutout.shape, np.min(dists), len(rows))

        attempts += 1

    t = Table(rows=rows, names=["IMAGE", "X", "Y", "SIZE", "MIN_DIST", "N_IN_CUTOUT", "RA", "DEC"])

    if output_filename is not None:
        t.write(output_filename)

    return t


def cutout_from_table(cutout_table, size=None):
    index = randint(0, len(cutout_table))
    row = cutout_table[index]

    size = row["SIZE"] if size is None else size
    image = galprime.select_image(row["IMAGE"])

    cutout = Cutout2D(image, (row["X"], row["Y"]), size=size)

    return cutout.data, row




__version__ = "1.0.2"

from .backgrounds import *
from .binning import *
from .config import *
from .data import *
from .densities import *
from .extraction import *
from .koe import *
from .masking import *
from .medians import *
from .modelling import *
from .profiles import *
from .sims import *

from . import plotting

def __main__():
    pass